# -*- coding: utf-8 -*-
import sys
import warnings
import wx
import dabo
if __name__ == "__main__":
	dabo.ui.loadUI("wx")
import dPemMixin as pm
import dIcons
from dabo.dLocalize import _
import dabo.dEvents as dEvents
from dabo.ui import makeDynamicProperty
from dabo.lib.utils import cleanMenuCaption

# wx constants for styles
NormalItemType = wx.ITEM_NORMAL
CheckItemType =  wx.ITEM_CHECK
RadioItemType = wx.ITEM_RADIO


class dMenu(pm.dPemMixin, wx.Menu):
	"""Creates a menu, which can contain submenus, menu items, 
	and separators.
	"""
	def __init__(self, parent=None, properties=None, attProperties=None, *args, **kwargs):
		self._baseClass = dMenu
		preClass = wx.Menu
		self.Parent = parent
		self._useMRU = self._extractKey(attProperties, "MRU", None)
		if self._useMRU is not None:
			self._useMRU = (self._useMRU == "True")
		else:
			self._useMRU = self._extractKey((properties, kwargs), "MRU", False)
		self._mruSeparator = None
		## pkm: When a dMenuItem is added to a dMenu, the wx functions only
		##      add the C++ portion, not the mixed-in dabo dMenuItem object.
		##      To work around this, we maintain an internal dictionary that
		##      maps the id of the wxMenuItem to the dMenuItem object.
		self._daboChildren = {}
		pm.dPemMixin.__init__(self, preClass, parent, properties, attProperties, *args, **kwargs)

		# Could be that we are used without a Dabo app object (not recommended,
		# but should be possible). So use the wx-method for getting the uiApp ref,
		# instead of the Dabo way of self.Application.uiApp
		uiApp = wx.GetApp()
		uiApp.Bind(wx.EVT_MENU_OPEN, self.__onWxMenuOpen)
		uiApp.Bind(wx.EVT_MENU_CLOSE, self.__onWxMenuClose)
		if self._useMRU:
			self.bindEvent(dEvents.MenuOpen, self._onMenuOpenMRU)

		
	def __onWxMenuOpen(self, evt):
		if not self:
			return
		cap = self.Caption
		try:
			if evt.GetMenu().Caption == cap:
				# Opening a single menu will trigger the wx event 
				# for every menu in the menubar.
				self.raiseEvent(dEvents.MenuOpen, evt)
			else:
				evt.Skip()
		except:
			evt.Skip()


	def __onWxMenuClose(self, evt):
		if not self:
			return
		cap = self.Caption
		try:
			if evt.GetMenu().Caption == cap:
				# Closing a single menu will trigger the wx event 
				# for every menu in the menubar.
				self.raiseEvent(dEvents.MenuClose, evt)
			else:
				evt.Skip()
		except:
			evt.Skip()


	def _onMenuOpenMRU(self, evt):
		if self.Application:
			self.Application.onMenuOpenMRU(self)
	
	
	def _initEvents(self):
		"""See self._setId(), which is where the binding of wxEvents needs to take 
		place.
		"""
		self.bindEvent(dEvents.MenuHighlight, self._onMenuHighlight)


	def _onMenuHighlight(self, evt):
		"""Note that this code is here in a dabo binding instead of in the wx binding
		because of the way we've worked around wx limitations: dMenu as a top-level
		menu in a menu bar doesn't send wx events.
		"""
		self._setDynamicEnabled()


	def _setDynamicEnabled(self):
		"""For each dMenuItem, set Enabled per the item's DynamicEnabled prop."""
		for item in self.Children:
			# separators haven't been abstracted yet, so there are still pure wx items.
			try:
				de = item.DynamicEnabled
			except:
				de = None
			if de is not None:
				if callable(de):
					item.Enabled = de()


	def __onWxMenuHighlight(self, evt):
		self.raiseEvent(dEvents.MenuHighlight)
		evt.Skip()


	def _getWxItem(self, wxFunc, dMenuItemInstance, pos=None):
		if pos is not None:
			wxItem = wxFunc(pos, dMenuItemInstance)
		else:
			wxItem = wxFunc(dMenuItemInstance)
		dMenuItemInstance.Parent = self
		self._daboChildren[wxItem.GetId()] = dMenuItemInstance
		if dMenuItemInstance.Icon and sys.platform.lower()[:3] == "win":
			wxItem.SetBitmap(dMenuItemInstance.Icon)
		return wxItem


	def appendItem(self, item):
		"""Insert a dMenuItem at the bottom of the menu."""
		wxItem = self._getWxItem(self.AppendItem, item)
		return item
		

	def insertItem(self, pos, item):
		"""Insert a dMenuItem before the specified position in the menu."""
		wxItem = self._getWxItem(self.InsertItem, item, pos)
		return item
		

	def prependItem(self, item):
		"""Insert a dMenuItem at the top of the menu."""
		wxItem = self._getWxItem(self.PrependItem, item)
		return item


	def appendMenu(self, menu):
		"""Insert a dMenu at the bottom of the menu."""
		wxMenuItem = self.AppendMenu(-1, menu.Caption, menu, help=menu.HelpText)
		menu._setId(wxMenuItem.GetId())
		menu.Parent = self
		self._daboChildren[wxMenuItem.GetId()] = menu
		return menu
		

	def insertMenu(self, pos, menu):
		"""Insert a dMenu before the specified position in the menu."""
		wxMenuItem = self.InsertMenu(pos, -1, menu.Caption, menu, help=menu.HelpText)
		menu._setId(wxMenuItem.GetId())
		menu.Parent = self
		self._daboChildren[wxMenuItem.GetId()] = menu
		return menu
		
		
	def prependMenu(self, menu):
		"""Insert a dMenu at the top of the menu."""
		wxMenuItem = self.PrependMenu(-1, menu.Caption, menu, help=menu.HelpText)
		menu._setId(wxMenuItem.GetId())
		menu.Parent = self
		self._daboChildren[wxMenuItem.GetId()] = menu
		return menu


	def appendSeparator(self):
		"""Insert a separator at the bottom of the menu."""
		return self.AppendSeparator()
		

	def insertSeparator(self, pos):
		"""Insert a separator before the specified position in the menu."""
		return self.InsertSeparator(pos)


	def prependSeparator(self):
		"""Insert a separator at the top of the menu."""
		return self.PrependSeparator()
	

	def append(self, caption, bindfunc=None, help="", bmp=None, picture=None,
			menutype="", *args, **kwargs):
		"""Append a dMenuItem with the specified properties.

		This is a convenient way to add a dMenuItem to a dMenu, give it a caption,
		help string, bitmap, and also bind it to a function, all in one call.
		
		NOTE: use of the bindfunc parameter is deprecated in version 0.8 and will be
		removed	in version 0.9. Send an OnHit parameter instead.

		Any additional keyword arguments passed will be interpreted as properties
		of the dMenuItem: if valid property names/values, the dMenuItem will take
		them on; if not valid, an exception will be raised.
		"""
		if picture is None:
			picture = bmp
		item = self._getItem(bindfunc, help, picture, menutype, *args, **kwargs)
		self.appendItem(item)
		item.Caption = caption
		return item
		
	
	def insert(self, pos, caption, bindfunc=None, help="", bmp=None, picture=None,
			menutype="", *args, **kwargs):
		"""Insert a dMenuItem at the given position with the specified properties.

		This is a convenient way to add a dMenuItem to a dMenu, give it a caption,
		help string, bitmap, and also bind it to a function, all in one call.

		NOTE: use of the bindfunc parameter is deprecated in version 0.8 and will be
		removed	in version 0.9. Send an OnHit parameter instead.

		Any additional keyword arguments passed will be interpreted as properties
		of the dMenuItem: if valid property names/values, the dMenuItem will take
		them on; if not valid, an exception will be raised.
		"""
		if picture is None:
			picture = bmp
		item = self._getItem(bindfunc, help, picture, menutype, *args, **kwargs)
		self.insertItem(pos, item)
		item.Caption = caption
		return item
		

	def prepend(self, caption, bindfunc=None, help="", bmp=None, picture=None,
			menutype="", *args, **kwargs):
		"""Prepend a dMenuItem with the specified properties.

		This is a convenient way to add a dMenuItem to a dMenu, give it a caption,
		help string, bitmap, and also bind it to a function, all in one call.

		NOTE: use of the bindfunc parameter is deprecated in version 0.8 and will be
		removed	in version 0.9. Send an OnHit parameter instead.

		Any additional keyword arguments passed will be interpreted as properties
		of the dMenuItem: if valid property names/values, the dMenuItem will take
		them on; if not valid, an exception will be raised.
		"""
		if picture is None:
			picture = bmp
		item = self._getItem(bindfunc, help, picture, menutype, *args, **kwargs)
		self.prependItem(item)
		item.Caption = caption
		return item
		

	def _resolveItem(self, capIdxOrItem):
		"""Returns the menu item specified by either its index or caption. In the
		case that an actual menu item is passed, simply returns that item.
		"""
		if isinstance(capIdxOrItem, basestring):
			ret = self.getItem(capIdxOrItem)
		elif isinstance(capIdxOrItem, int):
			ret = self.Children[capIdxOrItem]
		else:
			ret = capIdxOrItem
		return ret
		

	def remove(self, capIdxOrItem, release=True):
		"""Removes the specified item from the menu. You may specify the item by
		passing its index, its Caption, or by passing the item itself. If release is 
		True (the default), the item is destroyed as well. If release is False, a reference 
		to the object will be returned, and the caller is responsible for destroying it.
		"""
		item = self._resolveItem(capIdxOrItem)
		id_ = item.GetId()
		if self._daboChildren.has_key(id_):
			del self._daboChildren[id_]

		if wx.VERSION >= (2,7):
			# Needed to keep dPemMixin mixed-in in wxPython 2.8
			val = wx.Menu.RemoveItem(self, item)
			item.this.own(val.this.own())
			val.this.disown()
		else:
			self.RemoveItem(item)

		if release:
			item.Destroy()
			item = None
		return item


	def clear(self):
		"""Removes all items in this menu."""
		while self.Children:
			self.remove(0)
	
	
	def setItemCheck(self, itm, val):
		"""Pass a menu item and a boolean value, and the checked
		state of that menu item will be set accordingly.
		"""
		itm.Check(val)
		
		
	def setCheck(self, capIdxOrItem, unCheckOthers=True):
		"""When using checkmark-type menus, passing either the item
		itself, or the index or caption of the item you want checked to 
		this method will check that item. If unCheckOthers is True, non-
		matching items will be unchecked.
		"""
		target = self._resolveItem(capIdxOrItem)
		for itm in self.Children:
			if itm is target:
				try:
					itm.Checked =True
				except:
					pass
			else:
				if unCheckOthers:
					try:
						itm.Checked = False
					except:
						pass
	
	
	def clearChecks(self):
		"""Unchecks any checkmark-type menu items."""
		self.setCheck(None)
	
	
	def isItemChecked(self, capIdxOrItem):
		itm = self._resolveItem(capIdxOrItem)
		if itm is not None and itm.IsCheckable():
			ret = itm.Checked
		else:
			ret = None
		return ret
		
			
	def _getItem(self, bindfunc, help, icon, menutype, *args, **kwargs):
		if bindfunc is not None:
			warnings.warn(_("Deprecated; use 'OnHit=<func>' instead."), 
					DeprecationWarning, 1)
		itmtyp = self._getItemType(menutype)
		itmid = self._getItemID(menutype)
		if itmid != wx.ID_DEFAULT:
			kwargs["id"] = itmid
		cls = {NormalItemType: dabo.ui.dMenuItem,
				CheckItemType: dabo.ui.dCheckMenuItem,
				RadioItemType: dabo.ui.dRadioMenuItem}[itmtyp]
		itm = cls(self, HelpText=help, Icon=icon, kind=itmtyp, *args, **kwargs)
		if bindfunc:
			itm.bindEvent(dEvents.Hit, bindfunc)
		return itm


	def _getItemID(self,typ):
		typ = str(typ).lower()
		ret = wx.ID_DEFAULT
		if typ == "exit":
			ret = wx.ID_EXIT
		elif typ == "about":
			ret = wx.ID_ABOUT
		elif typ in ("pref", "prefs"):
			ret = wx.ID_PREFERENCES
		return ret
		

	def _getItemType(self, typ):
		typ = str(typ).lower()[:3]
		ret = NormalItemType
		if typ in ("che", "chk"):
			ret = CheckItemType
		elif typ == "rad":
			# Currently only implemented under Windows and GTK, 
			# use #if wxHAS_RADIO_MENU_ITEMS to test for 
			# availability of this feature.
			ret = RadioItemType
		return ret


	def _setId(self, id_):
		"""wxMenus don't have ids of their own - they only get set when the 
		menu gets added as a submenu - and then it becomes a wxMenuItem with a
		special submenu flag. This hook, called from append|insert|prependMenu(),
		allows the menu event bindings to take place.
		"""
		## MenuOpen and MenuClose don't appear to be working on Linux. Need
		## to test on Mac and Win.
		if self.Application is not None:
			# Set up a mechanism to catch menu events and re-raise Dabo events. 
			# If Application is None, however, this won't work because of wx 
			# limitations.
			self.Application.uiApp.Bind(wx.EVT_MENU_HIGHLIGHT,
					self.__onWxMenuHighlight, id=id_)
	
	
	def _itemByCaption(self, cap, returnPos=False):
		"""Common method for locating a menu item by its caption, ignoring
		all the 'special' characters for acceleration.
		"""
		cap = cleanMenuCaption(cap, "&_")
		for pos in xrange(self.GetMenuItemCount()):
			itm = self.FindItemByPosition(pos)
			itmCap = cleanMenuCaption(itm.GetLabel(), "&_")
			if itmCap == cap:
				if returnPos:
					return pos
				else:
					return self._daboChildren.get(itm.GetId(), None)
		return None


	def getItemIndex(self, caption):
		"""Returns the index of the item with the specified caption. If the item
		isn't found, None is returned.
		"""
		return self._itemByCaption(caption, True)
		

	def getItem(self, caption):
		"""Returns a reference to the menu item with the specified caption. If the item
		isn't found, None is returned.
		"""
		return self._itemByCaption(caption)


	def GetChildren(self):
		"""wx doesn't provide GetChildren() for menubars or menus, but dPemMixin
		calls it in _getChildren(). The Dabo developer wants the submenus and
		items in this menu, but is using the consistent Children property to 
		do it.
		"""
		children = self.GetMenuItems()
		daboChildren = [self._daboChildren.get(c.GetId(), c) for c in children]
		return daboChildren


	def _getCaption(self):
		try:
			v = self._caption
		except:
			v = self._caption = ""
		return v

	def _setCaption(self, val):
		if self._constructed():
			prnt = self.Parent
			if isinstance(prnt, wx.MenuBar):
				pos = prnt.FindMenu(self._caption)
				if pos >= 0:
					prnt.SetLabelTop(pos, val)
			else:
				isPopup = False
				while prnt:
					if isinstance(prnt, wx.MenuBar):
						isPopup = True
						break
					prnt = prnt.Parent
				if isPopup:
					# This is only needed for popups
					self.SetTitle(val)
			self._caption = val
		else:
			self._properties["Caption"] = val


	def _getEnabled(self):
		return self.IsEnabled()

	def _setEnabled(self, val):
		if self._constructed():
			self.Enable(bool(val))
		else:
			self._properties["Enabled"] = val


	def _getForm(self):
		if self.Parent:
			return self.Parent.Form
		return None


	def _getHelpText(self):
		try:
			v = self._helpText
		except AttributeError:
			v = self._helpText = ""
		return v

	def _setHelpText(self, val):
		self._helpText = val


	def _getParent(self):
		try:
			v = self._parent
		except AttributeError:
			v = self._parent = None
		return v

	def _setParent(self, val):
		self._parent = val

		
	Caption = property(_getCaption, _setCaption, None,
			_("Specifies the text of the menu."))

	Enabled = property(_getEnabled, _setEnabled, None,
			_("Specifies whether the menu can be interacted with."))

	Form = property(_getForm, None, None,
			_("Specifies the form that contains the menu."))

	HelpText = property(_getHelpText, _setHelpText, None,
			_("Specifies the help text associated with this menu. (str)"))

	Parent = property(_getParent, _setParent, None, 
			_("Specifies the parent menu or menubar."))


	DynamicCaption = makeDynamicProperty(Caption)
	DynamicEnabled = makeDynamicProperty(Enabled)
	DynamicHelpText = makeDynamicProperty(HelpText)

